import configparser
from datetime import datetime
import json
import os
import re
import shlex
import shutil
import stat
import subprocess as sp
import sys
import tempfile
from typing import Any, Dict, List, Optional, Union

import ConanTools

CONAN_CMD = os.environ.get("CT_CONAN_CMD", "conan")


def copytree(src: str, dst: str, symlinks: bool = True):
    """Custom copytree implementation that works with existing directories.

    This implementation has its roots in [1] and works around the problem that shutil.copytree
    before Python 3.8 does not work with existing directories. The alternatively recommended
    distutils.dir_util.copy_tree function supports this usecase but failed when when symlinks are
    enabled. Moreover, it is apparently not considered a public function [2].

    [1] https://stackoverflow.com/a/22331852
    [2] https://bugs.python.org/issue10948#msg337892
    """
    if not os.path.exists(dst):
        os.makedirs(dst)
        shutil.copystat(src, dst)
    lst = os.listdir(src)
    for item in lst:
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if symlinks and os.path.islink(s):
            if os.path.lexists(d):
                os.remove(d)
            os.symlink(os.readlink(s), d)
            try:
                st = os.lstat(s)
                mode = stat.S_IMODE(st.st_mode)
                os.lchmod(d, mode)
            except:
                pass  # lchmod not available
        elif os.path.isdir(s):
            copytree(s, d, symlinks)
        else:
            # Work around the fact that copy2 fails when the destination is not writeable.
            if os.path.exists(d) and not os.access(d, os.W_OK):
                os.chmod(d, stat.S_IWRITE)
            shutil.copy2(s, d)


def cmd_to_string(cmd: List[str]) -> str:
    return " ".join([shlex.quote(x) for x in cmd])


def create_stamp_file(path: str):
    with open(path, 'a'):
        pass


def _run_json(args: List[str]):
    try:
        tmpfile = tempfile.NamedTemporaryFile(suffix=".json", delete=False)
        tmpfile.close()
        sp.check_call([CONAN_CMD] + args + ["--json", tmpfile.name],
                      stdin=sp.DEVNULL, stdout=sp.DEVNULL, stderr=sp.DEVNULL)
        with open(tmpfile.name) as f:
            return json.load(f)
    finally:
        if tmpfile and os.path.exists(tmpfile.name):
            os.unlink(tmpfile.name)


# TODO use the check argument of sp.run (requires larger test updates)
def run(args: List[str], cwd: Optional[str] = None, stdout: Optional[int] = None,
        stderr: Optional[int] = None, check: bool = True, conan_cmd: str = CONAN_CMD):
    cmd = [conan_cmd] + args
    cmd_str = cmd_to_string(cmd)

    # ensure that the current working directory exists
    cwd = os.path.abspath(cwd if cwd is not None else os.getcwd())
    os.makedirs(cwd, exist_ok=True)

    # execute the actual command
    print("[{}] $ {}".format(cwd, cmd_str))
    sys.stdout.flush()
    result = sp.run(cmd, stdout=stdout, stderr=stderr, cwd=cwd)
    if stdout == sp.PIPE:
        result.stdout = result.stdout.decode().strip()
    if stderr == sp.PIPE:
        result.stderr = result.stderr.decode().strip()
    if check and result.returncode != 0:
        if stdout == sp.PIPE:
            print(result.stdout, file=sys.stdout)
        if stderr == sp.PIPE:
            print(result.stderr, file=sys.stderr)
        raise ValueError(
            "Executing command \"{}\" failed! (returncode={})".format(cmd_str, result.returncode))
    return result


def write_conan_sh_file(filedir: str, basename: str, args: List[str], cmd_cwd: Optional[str],
                        env: Optional[dict] = None, conan_cmd: str = CONAN_CMD):
    os.makedirs(filedir, exist_ok=True)
    filepath = os.path.join(filedir, "ct_{}.sh".format(basename))
    cmd_cwd = os.path.abspath(cmd_cwd if cmd_cwd is not None else os.getcwd())
    if env is None:
        env = os.environ
    with open(filepath, 'w') as f:
        f.write('#!/bin/sh\n')
        for k, v in env.items():
            f.write(cmd_to_string(['export', '{}={}'.format(k, v)]) + "\n")
        f.write(cmd_to_string(['cd', cmd_cwd]) + "\n")
        f.write(cmd_to_string([conan_cmd] + args) + "\n")

    # Ensure that the sh file is executable.
    # Source: https://stackoverflow.com/a/30463972
    mode = os.stat(filepath).st_mode
    mode |= (mode & 0o444) >> 2    # copy R bits to X
    os.chmod(filepath, mode)


def fmt_arg_list(values: Union[List[Optional[str]], Optional[str]], key: str) -> List[str]:
    """Generates an argument list by injecting option keys between the values.

    As value either one optional string or a list of optional strings are accepted. Hereby, None
    values are simply omitted and only the key is inserted into the argument list.
    """
    args = []
    if not isinstance(values, list):
        values = [values]
    for x in values:
        args.append(key)
        if x is not None:
            args.append(x)
    return args


def fmt_build_args(cmd: str, args: List[str], remote: Optional[str], profiles: List[str],
                   build: List[Optional[str]], options: Dict[str, str]) -> List[str]:
    profile_args = fmt_arg_list(profiles, "--profile")
    build_args = fmt_arg_list(build, "--build")
    remote_args = fmt_arg_list(remote or [], "--remote")
    option_args = fmt_arg_list(["{}={}".format(k, v) for k, v in options.items()], "-o")
    return [cmd] + args + profile_args + build_args + remote_args + option_args


def get_recipe_field(recipe_path, field_name, cwd=None):
    # make the recipe path absolute
    cwd = cwd or os.getcwd()
    if not os.path.isabs(recipe_path):
        recipe_path = os.path.normpath(os.path.join(cwd, recipe_path))

    if not os.path.exists(recipe_path):
        return None

    get_recipe_field._recipe_cache = getattr(get_recipe_field, "_recipe_cache", {})
    recipe = get_recipe_field._recipe_cache.setdefault(recipe_path, Recipe(recipe_path))
    return recipe.get_field(field_name)


class Reference():
    def __init__(self, name, version, user, channel):
        self._name = name
        self._version = version
        self._user = user
        self._channel = channel

    @classmethod
    def from_string(cls, ref: str) -> 'Reference':
        # FIXME support new conan center convention without user and channel
        reference_regex = re.compile(r'([\w\.\+\-]+)/([\w\.\+\-]+)@([\w\.\+\-]+)/([\w\.\+\-]+)')
        return cls(*reference_regex.match(ref).group(1, 2, 3, 4))

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def user(self):
        return self._user

    @property
    def channel(self):
        return self._channel

    def clone(self, name=None, version=None, user=None, channel=None):
        return Reference(name=name or self.name,
                         version=version or self.version,
                         user=user or self.user,
                         channel=channel or self.channel)

    def __str__(self):
        return "{}/{}@{}/{}".format(self.name, self.version, self.user, self.channel)

    def __repr__(self):
        return str(self)

    def in_local_cache(self):
        # check if the recipe is known locally
        result = run(["search", str(self)], check=False)
        if result.returncode == 0:
            return True
        return False

    def in_remote(self, remote):
        # check if the recipe is known on the remote
        result = run(["search", str(self), "--remote", remote], check=False)
        if result.returncode == 0:
            return True
        return False

    def get_creation_date(self, remote: Optional[str] = None) -> datetime:
        json_result = info(str(self), remote)
        return datetime.strptime(json_result['creation_date'], '%Y-%m-%d %H:%M:%S')

    def download_recipe(self, remote=None):
        remote_args = fmt_arg_list(remote or [], "--remote")
        run(["download", str(self), "--recipe"] + remote_args)

    def install(self, remote=None, profiles=[], build=["outdated"], options={}, cwd=None):
        args = fmt_build_args("install", [str(self)], remote=remote, profiles=profiles,
                              options=options, build=build)
        run(args, cwd=cwd)

    def set_remote(self, remote):
        run(["remote", "add_ref", str(self), remote])

    def create_alias(self, name=None, version=None, user=None, channel=None):
        alias_ref = self.clone(name=name, version=version, user=user, channel=channel)
        run(["alias", str(alias_ref), str(self)])
        return alias_ref

    def upload_all(self, remote):
        run(["upload", str(self), "--remote", remote, "--all", "-c"])


class PkgLayout():
    def root(self, recipe: 'Recipe') -> str:
        raise NotImplementedError

    def src_folder(self, recipe: 'Recipe') -> str:
        raise NotImplementedError

    def build_folder(self, recipe: 'Recipe') -> str:
        raise NotImplementedError

    def pkg_folder(self, recipe: 'Recipe') -> str:
        raise NotImplementedError


class RelativePkgLayout(PkgLayout):
    """Layout class that permits various relative package layouts.

    The default mode (i.e., no ``root`` specified) is to use the directory of the recipe as layout
    root. Subsequently, subdirectories relative to the recipe are used for all operations. The
    actual name of the subdirectories can be customized and an additional ``offset`` parameter can
    be used to further move the root relative to the recipe.

    Alternatively, when an explicit ``root`` has been specified, directories relative to this root
    are used. By default, the package name is used as ``offset`` to permit reusing the same layout
    instance accross multiple recipes.
    """
    def __init__(self, root: str = None, offset: str = None, src_dir: str = "_source",
                 build_dir: str = "_build", pkg_dir: str = "_install"):
        self._root = root
        self._offset = offset
        self._src_dir = src_dir
        self._build_dir = build_dir
        self._pkg_dir = pkg_dir

    def root(self, recipe: 'Recipe') -> str:
        if self._root is None:
            # No root has been defined, use the recipe path directory as root and apply the offset
            # if available.
            root = os.path.dirname(recipe.path)
            offset = self._offset or "."
        else:
            # Use the specified root and offset if available. Otherwise, fallback to the package
            # name as offset.
            root = self._root
            offset = self._offset or recipe.get_field("name")

        return os.path.normpath(os.path.join(root, offset))

    def src_folder(self, recipe: 'Recipe') -> str:
        if recipe.external_source:
            return os.path.join(self.root(recipe), self._src_dir)
        else:
            return os.path.dirname(recipe.path)

    def build_folder(self, recipe: 'Recipe') -> str:
        return os.path.join(self.root(recipe), self._build_dir)

    def pkg_folder(self, recipe: 'Recipe') -> str:
        return os.path.join(self.root(recipe), self._pkg_dir)


class TempPkgLayout(PkgLayout):
    def __init__(self, src_dir="_source", build_dir="_build",
                 pkg_dir="_install"):
        self._directories = {}
        self._src_dir = src_dir
        self._build_dir = build_dir
        self._pkg_dir = pkg_dir

    def __del__(self):
        for directory in self._directories.values():
            shutil.rmtree(directory)

    def root(self, recipe: 'Recipe') -> str:
        result = self._directories.get(recipe, False)
        if result:
            return result
        result = tempfile.mkdtemp(recipe.get_field("name"))
        self._directories[recipe] = result
        return result

    def src_folder(self, recipe: 'Recipe') -> str:
        if recipe.external_source:
            return os.path.join(self.root(recipe), self._src_dir)
        else:
            return os.path.dirname(recipe.path)

    def build_folder(self, recipe: 'Recipe') -> str:
        return os.path.join(self.root(recipe), self._build_dir)

    def pkg_folder(self, recipe: 'Recipe') -> str:
        return os.path.join(self.root(recipe), self._pkg_dir)


class Recipe():
    def __init__(self, path: str, external_source: bool = False,
                 layout: Optional[PkgLayout] = None, cwd: Optional[str] = None):
        # Make the recipe path absolute.
        cwd = cwd or os.getcwd()
        if not os.path.isabs(path):
            path = os.path.normpath(os.path.join(cwd, path))
        self._path = path

        # Ideally this property should be queryable from the recipe.
        # However at the moment I do not know how to do it.
        self._external_source = external_source

        self._layout = layout or RelativePkgLayout()

    @property
    def path(self):
        return self._path

    @property
    def external_source(self):
        return self._external_source

    def get_field(self, field_name: str, default: Any = None):
        return inspect(self.path, attribute=field_name, default=default)

    def reference(self, user: str, channel: str, name=None, version=None):
        name = name or self.get_field("name")
        version = version or self.get_field("version")
        return Reference(name=name, version=version, user=user, channel=channel)

    def export(self, user, channel, name=None, version=None):
        ref = self.reference(name=name, version=version, user=user, channel=channel)
        run(["export", self.path, str(ref)])
        return ref

    def create(self, user, channel, name=None, version=None, remote=None,
               profiles=[], options={}, build=["outdated"], cwd=None):
        ref = self.reference(name=name, version=version, user=user, channel=channel)
        args = fmt_build_args("create", [self.path, str(ref)], remote=remote, profiles=profiles,
                              options=options, build=build)
        run(args, cwd=cwd)
        return ref

    def create_local(self, user, channel, name=None, version=None, remote=None,
                     profiles=[], options={}, build=["outdated"], layout=None,
                     src_folder=None, build_folder=None, pkg_folder=None, add_script=False):
        self.install(layout=layout, build_folder=build_folder, profiles=profiles, options=options,
                     build=build, remote=remote, add_script=add_script)
        if self.external_source:
            self.source(layout=layout, src_folder=src_folder, build_folder=build_folder,
                        add_script=add_script)
        self.build(layout=layout, src_folder=src_folder, build_folder=build_folder,
                   pkg_folder=pkg_folder, add_script=add_script)
        self.package(layout=layout, src_folder=src_folder, build_folder=build_folder,
                     pkg_folder=pkg_folder, add_script=add_script)
        return self.export_pkg(user=user, channel=channel, name=name, version=version,
                               profiles=profiles, options=options, layout=layout,
                               pkg_folder=pkg_folder, add_script=add_script)

    def install(self, layout=None, build_folder=None, profiles=[], options={}, build=["outdated"],
                remote=None, add_script=False):
        layout = layout or self._layout
        build_folder = build_folder or layout.build_folder(self)
        args = fmt_build_args("install", [self.path], remote=remote, profiles=profiles,
                              options=options, build=build)
        if add_script:
            write_conan_sh_file(layout.root(self), 'install', args, build_folder)
        run(args, cwd=build_folder)

    def source(self, layout=None, src_folder=None, build_folder=None, add_script=False):
        layout = layout or self._layout
        src_folder = src_folder or layout.src_folder(self)
        stamp_file = os.path.join(src_folder, ".ct_source_finished")
        if os.path.isfile(stamp_file):
            return
        # Delete the source folder if it already exists.
        shutil.rmtree(src_folder, ignore_errors=True)
        build_folder = build_folder or layout.build_folder(self)
        args = ["source", self.path, "--source-folder=" + src_folder]
        if add_script:
            write_conan_sh_file(layout.root(self), 'source', args, build_folder)
        run(args, cwd=build_folder)
        # Create the stamp file after successfully executing conan source.
        create_stamp_file(stamp_file)

    def build(self, layout=None, src_folder=None, build_folder=None, pkg_folder=None,
              add_script=False):
        layout = layout or self._layout
        src_folder = src_folder or layout.src_folder(self)
        build_folder = build_folder or layout.build_folder(self)
        pkg_folder = pkg_folder or layout.pkg_folder(self)
        args = ["build", self.path, "--source-folder=" + src_folder,
                "--package-folder=" + pkg_folder]
        if src_folder != build_folder and self.get_field("no_copy_source", False) is False:
            # Unlike "conan create", "conan build" does not copy the source folder to the build
            # folder automatically. We have to perform this copy because recipes depend on it.
            copytree(src_folder, build_folder)
        if add_script:
            write_conan_sh_file(layout.root(self), 'build', args, build_folder)
        run(args, cwd=build_folder)

    def package(self, layout=None, src_folder=None, build_folder=None, pkg_folder=None,
                add_script=False):
        layout = layout or self._layout
        src_folder = src_folder or layout.src_folder(self)
        build_folder = build_folder or layout.build_folder(self)
        pkg_folder = pkg_folder or layout.pkg_folder(self)
        args = ["package", self.path, "--source-folder=" + src_folder,
                "--package-folder=" + pkg_folder]
        if add_script:
            write_conan_sh_file(layout.root(self), 'package', args, build_folder)
        run(args, cwd=build_folder)

    def export_pkg(self, user: str, channel: str, name: Optional[str] = None,
                   version: Optional[str] = None, force: bool = True, profiles: List[str] = [],
                   options: Dict[str, str] = {}, layout: Optional[PkgLayout] = None,
                   pkg_folder: Optional[str] = None, cwd: Optional[str] = None,
                   add_script: bool = False):
        layout = layout or self._layout
        pkg_folder = pkg_folder or layout.pkg_folder(self)
        ref = self.reference(name=name, version=version, user=user, channel=channel)
        args = fmt_build_args("export-pkg", [self.path, str(ref), "--package-folder=" + pkg_folder],
                              remote=None, profiles=profiles, options=options, build=[])
        if force:
            args.append("--force")
        if add_script:
            write_conan_sh_file(layout.root(self), 'export-pkg', args, cwd)
        run(args, cwd=cwd)
        return ref


class Workspace():
    def __init__(self, recipes: List[Recipe]):
        self._recipes = recipes

    def references(self, user: str, channel: str):
        return [recipe.reference(user=user, channel=channel) for recipe in self._recipes]

    def install(self, user: str, channel: str, ws_build_folder: Optional[str] = None,
                profiles: List[str] = [], options: Dict[str, str] = {},
                build: List[Optional[str]] = ["outdated"], remote: Optional[str] = None,
                add_script: bool = False):
        config = configparser.ConfigParser(allow_no_value=True)
        config.optionxform = str
        for recipe in self._recipes:
                ref = recipe.reference(user, channel)
                layout = recipe._layout  # FIXME add accessor
                config["{}:build_folder".format(ref)] = {layout.build_folder(recipe): None}
                config["{}:source_folder".format(ref)] = {layout.src_folder(recipe): None}

        ws_build_folder = ws_build_folder or os.getcwd()
        os.makedirs(ws_build_folder, exist_ok=True)
        layout_file = os.path.join(ws_build_folder, "layout.txt")
        with open(layout_file, 'w') as f:
            config.write(f)

        ws_file = os.path.join(ws_build_folder, "ws.yml")
        with open(ws_file, 'w') as f:
            f.write("editables:\n")
            for recipe in self._recipes:
                f.write("    {}:\n".format(recipe.reference(user, channel)))
                f.write("        path: {}\n".format(os.path.dirname(recipe.path)))
            f.write("layout: layout.txt\n")
            # f.write("workspace_generator: cmake\n")
            f.write("root:\n")
            for recipe in self._recipes:
                f.write("  - {}\n".format(recipe.reference(user, channel)))

        args = ["workspace"]
        args += fmt_build_args("install", [ws_file], remote=remote, profiles=profiles,
                               options=options, build=build)
        if add_script:
            write_conan_sh_file(ws_build_folder, 'ws-install', args, ws_build_folder)
        run(args, cwd=ws_build_folder)

    def source(self, add_script: bool = False):
        for recipe in self._recipes:
            if recipe.external_source:
                recipe.source(add_script=add_script)

    def create_local(self, user: str, channel: str, ws_build_folder: Optional[str] = None,
                     profiles: List[str] = [], options: Dict[str, str] = {},
                     build: List[Optional[str]] = ["outdated"], remote: Optional[str] = None,
                     pkg_folder: Optional[str] = None, pkg_folder_override: Dict[Recipe, str] = {},
                     add_script: bool = False):
        self.install(user, channel, ws_build_folder=ws_build_folder, profiles=profiles,
                     options=options, build=build, remote=remote, add_script=add_script)
        self.source(add_script=add_script)
        # FIXME extract dependency information for correct build ordering
        for recipe in self._recipes:
            recipe_pkg_folder = pkg_folder_override.get(recipe, pkg_folder)
            recipe.build(pkg_folder=recipe_pkg_folder, add_script=add_script)
            recipe.package(pkg_folder=recipe_pkg_folder, add_script=add_script)
            if recipe_pkg_folder is None:
                recipe.export_pkg(user=user, channel=channel, profiles=profiles,
                                  options=options, add_script=add_script)


def search(pattern: str = "*", remote: Optional[str] = None) -> List[Reference]:
    json_result = _run_json(["search", pattern] + fmt_arg_list(remote or [], "--remote"))
    # FIXME check the json_result['error'] field
    # FIXME support multiple remotes
    assert len(json_result['results']) == 1
    assert json_result['results'][0]['remote'] == remote
    return [Reference.from_string(x['recipe']['id']) for x in json_result['results'][0]['items']]


def info(path_or_ref: str, remote: Optional[str] = None):
    # NOTE: Conan implicitely downloads the recipe if it is not available locally.
    json_result = _run_json(["info", path_or_ref] + fmt_arg_list(remote or [], "--remote"))
    assert len(json_result) == 1
    return json_result[0]


def inspect(path_or_ref: str, attribute: Optional[str] = None,
            default: Any = None,
            remote: Optional[str] = None) -> Union[dict, Any]:
    json_result = _run_json(["inspect", path_or_ref] +
                            fmt_arg_list(attribute or [], "--attribute") +
                            fmt_arg_list(remote or [], "--remote"))
    if attribute:
        # conan returns an empty string if the attribute is not defined.
        # We replace this sentinel with the user defined default value.
        res = json_result[attribute]
        if res == '':
            return default
        return res
    return json_result



import os
import string
from typing import Dict, List, Optional

import ConanTools.Conan as Conan
import ConanTools.Repack
import ConanTools.Version


def slug(input: Optional[str]) -> Optional[str]:
    """Creates a lowercase alphanumeric version of the input with '-' for all other characters.

    Additionally, all '-' characters are stripped from the start and end of the result.
    The approach is similar to the one Gitlab CI uses for _SLUG variables (see
    `CI_COMMIT_REF_SLUG <https://docs.gitlab.com/ee/ci/variables/predefined_variables.html>`_).

    :param input: Optional input string.
    :returns: Slug version of the input string or None if input is None.
    """
    if input is None:
        return None
    whitelist = set(string.ascii_letters + string.digits)

    def sanitize_char(ch):
        if ch in whitelist:
            return ch.lower()
        return '-'

    return ''.join([sanitize_char(ch) for ch in input]).strip('-')


def env_flag(name: str, default: bool = False) -> bool:
    """Queries an environment variable and converts the value to a bool.

    :param name: Name of the environment variable.
    :param default: Default value that is returned when the variable is not defined.
    :returns: Boolean interpretation of the value.
    """
    res = os.environ.get(name, default)
    if isinstance(res, bool):
        return res
    res = res.lower()
    if res == "0" or res == "false" or res == "off":
        return False
    return True


def pkg_create(recipe: Conan.Recipe, user: str, channel: str, name: Optional[str] = None,
               version: Optional[str] = None, remote: Optional[str] = None,
               profiles: List[Optional[str]] = ["outdated"], options: Dict[str, str] = {},
               build: Optional[List[str]] = None, cwd: Optional[str] = None,
               layout: Optional['Conan.PkgLayout'] = None, create_local: Optional[bool] = None):
    """Creates a package from the recipe using either the local or cache-based workflow.

    The ``CT_CREATE_LOCAL`` environment variable is used to enable the local instead of the
    cache-based flow. By default, the local flow builds into fixed directories next to the recipe
    which is more comfortable during development and also better suited for build caching.
    """
    if create_local is None:
        create_local = env_flag("CT_CREATE_LOCAL")
    if create_local:
        recipe.create_local(user=user, channel=channel, name=name, version=version, remote=remote,
                            profiles=profiles, options=options, build=build, layout=layout,
                            add_script=True)
    else:
        recipe.create(user=user, channel=channel, name=name, version=version, remote=remote,
                      profiles=profiles, options=options, build=build, cwd=cwd)


def pkg_import(recipe: Conan.Recipe, user: str, channel: str, name: Optional[str] = None,
               version: Optional[str] = None, remote: Optional[str] = None,
               profiles: List[str] = [], options: Dict[str, str] = {},
               build: List[Optional[str]] = ["outdated"], pkg_folder: Optional[str] = None,
               enable_subpackages: Optional[bool] = None):
    """Imports the package content, after building it if necessary, into the pkg_folder.

    By default, subpackages are built using the local flow and directly use the specified
    pkg_folder. This mode is similar to, for example, a superbuild using cmake. Alternatively,
    if the ``CT_ENABLE_SUBPACKAGES`` environment variable is defined, each subpackage is
    created individually and gets subsequently imported into the pkg_folder.
    """
    if enable_subpackages is None:
        enable_subpackages = env_flag("CT_ENABLE_SUBPACKAGES")

    reference = recipe.reference(user=user, channel=channel, name=name, version=version)

    # qualify the options with the package name
    full_opt = {}
    for k, v in options.items():
        if ":" in k:
            full_opt[k] = v
        else:
            full_opt["{}:{}".format(reference.name, k)] = v

    if not enable_subpackages:
        # Build package with the local flow but skip real package creation. Install
        # directly into the pkg_folder instead.
        recipe.install(profiles=profiles, options=full_opt, build=build, remote=remote,
                       add_script=True)
        if recipe.external_source:
            recipe.source(add_script=True)
        recipe.build(pkg_folder=pkg_folder, add_script=True)
        recipe.package(pkg_folder=pkg_folder, add_script=True)
        return

    # Try to import an already existing package but without building it.
    importFile = ConanTools.Repack.ConanImportTxtFile()
    importFile.add_package(reference)
    try:
        importFile.install(remote=remote, profiles=profiles, options=full_opt, build=[],
                           cwd=pkg_folder)
        return
    except ValueError:
        pass

    # Build the package using the local or cache-based workflow and then import the content.
    pkg_create(recipe=recipe, user=user, channel=channel, name=name, version=version, remote=remote,
               profiles=profiles, options=full_opt, build=build)
    importFile.install(remote=remote, profiles=profiles, options=full_opt, build=[], cwd=pkg_folder)


def ws_import(ws: Conan.Workspace, user: str, channel: str, name: Optional[str] = None,
              version: Optional[str] = None, remote: Optional[str] = None,
              profiles: List[str] = [], options: Dict[str, str] = {},
              build: List[Optional[str]] = ["outdated"], pkg_folder: Optional[str] = None,
              enable_subpackages: Optional[bool] = None, cwd=None,
              pkg_folder_override: Dict[Conan.Recipe, str] = {}):
    """Imports the workspace content, after building it if necessary, into the pkg_folder.

    By default, subpackages are built using the local flow and directly use the specified
    pkg_folder. This mode is similar to, for example, a superbuild using cmake. Alternatively,
    if the ``CT_ENABLE_SUBPACKAGES`` environment variable is defined, each subpackage is
    created individually and gets subsequently imported into the pkg_folder.
    """
    pkg_folder = pkg_folder or os.getcwd()
    if enable_subpackages is None:
        enable_subpackages = env_flag("CT_ENABLE_SUBPACKAGES")

    if not enable_subpackages:
        # Build workspace with the local flow but skip real package creation. Install
        # directly into the pkg_folder instead.
        ws.create_local(user, channel, ws_build_folder=cwd, profiles=profiles, options=options,
                        build=build, remote=remote, pkg_folder=pkg_folder,
                        pkg_folder_override=pkg_folder_override, add_script=True)
        return

    assert False


def write_helper_scripts(filedir: str, recipe_path: str, src_folder: str = None,
                         build_folder: str = None, pkg_folder: str = None):
    """Generate helper shell scripts for executing the build and package stage.
    """
    Conan.write_conan_sh_file(filedir, "build",
                              ["build", recipe_path, "--source-folder=" + src_folder,
                               "--package-folder=" + pkg_folder], build_folder)
    Conan.write_conan_sh_file(filedir, "package",
                              ["package", recipe_path, "--package-folder=" + pkg_folder],
                              build_folder)



import configparser
import glob
import tempfile
from typing import List
import os

import ConanTools.Conan as Conan


class ConanImportTxtFile:
    def __init__(self, file_name=None, cwd=None):
        self._package_ids = {}
        self._file_name = file_name
        self._delete = False

        # create a temporary file name if needed
        if self._file_name is None:
            tmpfile = tempfile.NamedTemporaryFile(suffix=".txt", dir=cwd, delete=False)
            tmpfile.close()
            self._file_name = tmpfile.name
            self._delete = True

    def __del__(self):
        if self._delete and os.path.exists(self._file_name):
            os.unlink(self._file_name)

    def add_package_string(self, name, refstring: str):
        self._package_ids[name] = refstring

    def add_package(self, ref: Conan.Reference):
        self._package_ids[ref.name] = str(ref)

    def add_packages(self, refs: List[Conan.Reference]):
        for ref in refs:
            self.add_package(ref)

    def install(self, remote=None, profiles=[], options={}, build=["outdated"], cwd=None):
        # write a conanfile in txt format with the package ids the imports
        config = configparser.ConfigParser(allow_no_value=True)
        config.optionxform = str
        config["requires"] = {x: None for x in self._package_ids.values()}
        config["imports"] = {
            "., * -> . @ root_package={}".format(x): None for x in self._package_ids.keys()
        }
        with open(self._file_name, 'w') as configfile:
            config.write(configfile)

        args = Conan.fmt_build_args("install", [self._file_name], remote=remote, profiles=profiles,
                                    options=options, build=build)
        Conan.run(args, cwd=cwd)

        # remove conan packaging metadata files
        cwd = os.path.abspath(cwd if cwd is not None else os.getcwd())
        files = glob.glob(os.path.join(cwd, "conan*"))
        files += glob.glob(os.path.join(cwd, "graph_info.json"))
        for f in files:
            os.remove(f)


def extend_profile(inpath, outpath, build_requires):
    config = configparser.ConfigParser(allow_no_value=True)
    config.optionxform = str
    config.read([inpath])
    for x in build_requires:
        config["build_requires"][x] = None
    with open(outpath, 'w') as configfile:
        config.write(configfile)



"""Support module for querying git.

This module provides a few simple methods to deduce information like the current revision, the name
of the current tag, or potential branch names from a git repository. We call git using a subprocess
to the deduce the needed information which means that a shell installation of git is required to
use this module.
"""
import os
from pathlib import Path
from subprocess import run, PIPE, DEVNULL
from typing import List, Optional


def revision(cwd: Optional[str] = None) -> Optional[str]:
    res = run(["git", "rev-parse", "HEAD"],
              stdin=DEVNULL, stdout=PIPE, stderr=DEVNULL, universal_newlines=True, cwd=cwd)
    if res.returncode == 0:
        return res.stdout.strip()
    return None


def branches(rev: Optional[str] = None, cwd: Optional[str] = None) -> List[str]:
    current_sha = rev or revision(cwd=cwd)
    # Get a list of all heads and their SHAs.
    refs = run(["git", "for-each-ref", "--format=%(objectname) %(refname:short)", "refs/heads"],
               stdin=DEVNULL, stdout=PIPE, stderr=DEVNULL, universal_newlines=True, cwd=cwd,
               check=True).stdout.strip()
    if refs != "":
        refs = [line.split(' ', 1) for line in refs.replace('\r', '').split('\n')]
        return [name for sha, name in refs if sha == current_sha]
    # Fallback to remote refs if no heads exist.
    refs = run(["git", "for-each-ref", "--format=%(objectname) %(refname)", "refs/remotes"],
               stdin=DEVNULL, stdout=PIPE, stderr=DEVNULL, universal_newlines=True, cwd=cwd,
               check=True).stdout.strip()
    if refs != "":
        refs = [line.split(' ', 1) for line in refs.replace('\r', '').split('\n')]
        # Strip the first 3 components from the ref to get the branch name
        # (newer git versions have "--format=%(objectname) %(refname:lstrip=3)")
        return [os.path.join(*Path(name).parts[3:]) for sha, name in refs if sha == current_sha]
    return None


def branch(cwd: Optional[str] = None) -> Optional[str]:
    res = branches(cwd=cwd)
    if res:
        return res[0]
    return None


def describe(cwd: Optional[str] = None) -> Optional[str]:
    res = run(["git", "describe", "--tags", "--abbrev=40", "--always"],
              stdin=DEVNULL, stdout=PIPE, stderr=DEVNULL, universal_newlines=True, cwd=cwd)
    if res.returncode == 0:
        return res.stdout.strip()
    return None


def is_repository(cwd: Optional[str] = None) -> bool:
    return run(["git", "status"],
               stdin=DEVNULL, stdout=DEVNULL, stderr=DEVNULL, cwd=cwd).returncode == 0


def tag(cwd: Optional[str] = None) -> Optional[str]:
    res = run(["git", "describe", "--exact-match", "--tags"],
              stdin=DEVNULL, stdout=PIPE, stderr=DEVNULL, universal_newlines=True, cwd=cwd)
    if res.returncode == 0:
        return res.stdout.strip()
    return None



"""Support module that permits to derive the version of the ConanTools package from git.

The idea of this module is to generate sane package versions also for non-release configurations
based on git describe. The overall approach is similar to packages like
`setuptools_scm <https://github.com/pypa/setuptools_scm>`_ and
`dunamai <https://github.com/mtkennerly/dunamai>`_. However, given that we want to distribute
ConanTools also as conan package, having external dependencies is not really appealing.
"""
import os
from ConanTools import Git
from typing import Optional


# The following version string is only relevant when the repository gets tagged, when the package
# is deployed (e.g., pypi), or before the first tag is created in the repository.
# -> Update it to the release version before cutting the release.
__version_string__ = '0.4.1'


def is_release(cwd: Optional[str] = None) -> bool:
    """Returns True if the folder is a git repository and on a tag or not a git repository at all.
    """
    if not Git.is_repository(cwd):
        return True
    if Git.tag(cwd) is not None:
        return True
    return False


def _format_git_version(fallback: str, cwd: Optional[str], digits: int,
                        mod_sep: str, metadata_sep: str) -> str:
    if cwd == "":
        cwd = os.path.dirname(os.path.abspath(__file__))
    if is_release(cwd):
        return fallback
    desc_str = Git.describe(cwd)
    parts = desc_str.split("-")
    if len(parts) == 1:
        # only relevant when no tag was found and only the SHA has been returned
        parts = [fallback, "dev0", "g" + desc_str]
    else:
        parts[1] = "post" + parts[1]
    res = parts[0] + mod_sep + parts[1]
    if digits == 0:
        return res
    return res + metadata_sep + parts[2][0:digits + 1]


# https://www.python.org/dev/peps/pep-0440
def pep440(fallback: str = __version_string__,
           cwd: Optional[str] = "", digits: int = 10) -> str:
    """Deduces the package version via git describe and returns it in PEP440 format.

    :param fallback: The version string that is returned when no git based version can be derived.
    :param cwd: The directory where git should be invoked. (None -> current dir, "" -> package dir)
    :param digits: Number of SHA digits that should be appended as metadata. [0-40]
    :returns: The resulting version string formatted according to PEP440.
    """
    return _format_git_version(fallback, cwd, digits, ".", "+")


# https://semver.org/spec/v2.0.0.html
def semantic(fallback: str = __version_string__,
             cwd: Optional[str] = "", digits: int = 10) -> str:
    """Deduces the package version via git describe and returns it in semantic versioning format.

    :param fallback: The version string that is returned when no git based version can be derived.
    :param cwd: The directory where git should be invoked. (None -> current dir, "" -> package dir)
    :param digits: Number of SHA digits that should be appended as metadata. [0-40]
    :returns: The resulting version string formatted according to semantic versioning.
    """
    return _format_git_version(fallback, cwd, digits, "-", "+")



import argparse
import inspect
import os


def reach(var_name, function_name=None):
    """Helper to search for a local variable by traversing the call stack.
    """
    for f in reversed(inspect.stack()):
        if function_name and f.function != function_name:
            continue
        if var_name in f[0].f_locals:
            return f[0].f_locals[var_name]
    return None


def get_cl_profiles():
    """ HACK Determines the profiles which are specified when invoking conan.

    This function determines the initial working directory by inspecting the
    call stack and reparses the arguments to extract paths to the profile
    files.
    """
    initial_cwd = reach("current_dir", function_name="main")
    if initial_cwd is None:
        return []

    parser = argparse.ArgumentParser(add_help=False, allow_abbrev=False)
    parser.add_argument("-pr", "--profile", action='append', dest='profiles', default=[])
    args, _ = parser.parse_known_args()
    profiles = []
    for x in args.profiles:
        if not os.path.isabs(x):
            fullpath = os.path.normpath(os.path.join(initial_cwd, x))
            if os.path.exists(fullpath):
                profiles.append(fullpath)
                continue
        profiles.append(x)
    return profiles


def get_cl_build_flags():
    """ HACK Determine the build flags which are specified when invoking conan.

    This function reparses the arguments to extract the build flags.
    """
    parser = argparse.ArgumentParser(add_help=False, allow_abbrev=False)
    parser.add_argument("-b", "--build", action='append', dest='build', nargs="?", default=[])
    args, _ = parser.parse_known_args()
    return args.build



import os
from setuptools import setup
import sys

script_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, script_dir)
from ConanTools import Version  # noqa


def readme():
    with open(os.path.join(script_dir, 'README.rst')) as f:
        return f.read()


pkgname = 'ConanTools'
pkgversion = Version.pep440()

# make the sphinx command available if sphinx is installed
try:
    from sphinx.setup_command import BuildDoc
    cmdclass = {'sphinx': BuildDoc}
except ImportError:
    cmdclass = {}

setup(author='Mario Werner',
      author_email='nioshd@gmail.com',
      classifiers=[  # https://pypi.org/classifiers/
          'Development Status :: 2 - Pre-Alpha',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.5',
          'Topic :: Software Development :: Build Tools',
          'Topic :: Utilities',
      ],
      cmdclass=cmdclass,
      command_options={
          'sphinx': {
              'build_dir': ('setup.py', 'doc/_build'),
              'project': ('setup.py', pkgname),
              'release': ('setup.py', pkgversion),
              'source_dir': ('setup.py', 'doc'),
              'version': ('setup.py', pkgversion),
          }
      },
      description='Helpers and tools that make working with conan more convenient.',
      extras_require={
          'documentation': ['sphinx', 'sphinx-autodoc-typehints'],
      },
      license='MIT',
      long_description=readme(),
      name=pkgname,
      packages=[pkgname],
      python_requires='>=3.5',
      setup_requires=['pytest-runner'],
      tests_require=['pytest', 'pytest-mock'],
      url='https://github.com/niosHD/ConanTools',
      version=pkgversion,
      zip_safe=False,
      )
